"""Helper functions for data io operations."""

import os
from typing import Dict

import imageio
import tqdm

from redbrick_sagemaker.logging import logger


def download_images_from_remote(cache_dir: str, prefix: str, tasks: Dict):
    """
    Download locally preparing for s3 upload for inference.

    Parameters
    -------------
    cache_dir: str
        Where images get cached

    prefix: str
        Sub-folder

    tasks: Dict
        RedBrick AI tasks.
    """
    os.makedirs(os.path.join(cache_dir, prefix), exist_ok=True)
    for task in tqdm.tqdm(tasks):
        # Download, if not already do
        local_path = os.path.join(cache_dir, prefix, str(task["taskId"]) + ".png")
        if os.path.exists(local_path):
            continue

        try:
            image = imageio.imread(task["itemsPresigned"][0])
        except (ValueError, OSError):
            logger.warning("Failed to download %s", task["items"])

        # Write to disk
        try:
            imageio.imwrite(
                os.path.join(local_path),
                image,
            )
        except (ValueError, OSError):
            logger.warning("Failed to download %s", task["items"][0])



"""Sagemaker related helper functions."""


from typing import Any, Optional


def construct_training_job(
    training_image: Any,
    train_uri: str,
    validation_uri: str,
    bucket: str,
    num_classes: int,
    num_training_samples: int,
    sagemaker_role: str,
    max_epochs: Optional[int] = 200,
    instance_type: Optional[str] = "ml.p2.xlarge",
):
    """Construct a training job object to configure sagemaker."""
    training_job_definition = {
        "AlgorithmSpecification": {
            "TrainingImage": training_image,
            "TrainingInputMode": "File",
        },
        "InputDataConfig": [
            {
                "ChannelName": "train",
                "CompressionType": "None",
                "ContentType": "application/x-recordio",
                "DataSource": {
                    "S3DataSource": {
                        "S3DataDistributionType": "FullyReplicated",
                        "S3DataType": "S3Prefix",
                        "S3Uri": train_uri,
                    }
                },
            },
            {
                "ChannelName": "validation",
                "CompressionType": "None",
                "ContentType": "application/x-recordio",
                "DataSource": {
                    "S3DataSource": {
                        "S3DataDistributionType": "FullyReplicated",
                        "S3DataType": "S3Prefix",
                        "S3Uri": validation_uri,
                    }
                },
            },
        ],
        "OutputDataConfig": {"S3OutputPath": f"s3://{bucket}/output"},
        "ResourceConfig": {
            "InstanceCount": 1,
            "InstanceType": instance_type,
            "VolumeSizeInGB": 20,
        },
        "RoleArn": sagemaker_role,
        "StaticHyperParameters": {
            "num_classes": str(num_classes),
            "num_training_samples": str(num_training_samples),
            "epochs": str(max_epochs),
            "use_pretrained_model": "1",
            "early_stopping": "True",
        },
        "StoppingCondition": {"MaxRuntimeInSeconds": 43200},
    }
    return training_job_definition


def construct_tuning_job(
    num_training_samples: int,
    max_jobs: Optional[int] = 10,
):
    """Construct tuning job to configure sagemaker."""
    # dynamically define batch size bounds
    min_batch = "8"
    max_batch = "32"
    if num_training_samples < 200:
        max_batch = "64"
    if num_training_samples < 80:
        max_batch = "16"
    if num_training_samples <= 50:
        raise Exception("You need at least 70 labeled images to train!")

    tuning_job_config = {
        "TrainingJobEarlyStoppingType": "Auto",
        "ParameterRanges": {
            "CategoricalParameterRanges": [
                {
                    "Name": "optimizer",
                    "Values": ["sgd", "adam", "rmsprop", "adadelta"],
                }
            ],
            "ContinuousParameterRanges": [
                {
                    "MinValue": "1e-6",
                    "MaxValue": "0.5",
                    "Name": "learning_rate",
                },
                {"MinValue": "0", "MaxValue": "0.999", "Name": "momentum"},
                {"MinValue": "0", "MaxValue": "0.999", "Name": "weight_decay"},
            ],
            "IntegerParameterRanges": [
                {
                    "MaxValue": max_batch,
                    "MinValue": min_batch,
                    "Name": "mini_batch_size",
                }
            ],
        },
        "ResourceLimits": {
            "MaxNumberOfTrainingJobs": max_jobs,
            "MaxParallelTrainingJobs": 1,
        },
        "Strategy": "Bayesian",
        "HyperParameterTuningJobObjective": {
            "MetricName": "validation:mAP",
            "Type": "Maximize",
        },
    }
    return tuning_job_config


def construct_warm_start_config(old_job_name: str):
    """Construct a warm start config."""
    warm_start_config = {
        "ParentHyperParameterTuningJobs": [
            {"HyperParameterTuningJobName": old_job_name}
        ],
        "WarmStartType": "TransferLearning",
    }
    return warm_start_config



"""Utility functions for bounding box jobs."""
import os
import json

from typing import Dict, Tuple

import numpy as np
import tqdm
import imageio

from redbrick_sagemaker.utils.redbrick import create_taxonomy_map
from .im2rec import run_im2rec


def convert_sagemaker_rbai_bbox(  # pylint: disable=too-many-locals
    prediction_dir: str,
    tasks: Dict,
    taxonomy: Dict[int, str],
    thresh: float = 0.4,
):
    """
    Convert a sagemaker detection to rbai task.

    Parameters
    ------------
    prediction_dir: str
        Local path where the prediction files are stored.
        The filenames will be task_id.png.out

    tasks: List[Dict]
        RBAI task objects for which the predictions are stored
        in prediction_dir

    taxonomy: Dict[int, str]
        maps from class_id to class_name

    thresh: float = 0.4
        confidence below this threshold will be ignored.

    Returns
    ------------
    labeled_tasks: List[Dict]
        Returns all the tasks with labels.
    """
    labeled_tasks = []
    for task in tasks:
        task_id = task["taskId"]
        prediction_file = os.path.join(prediction_dir, f"{task_id}.png.out")

        # read single pred file
        with open(
            prediction_file,
            "r",
            encoding="Utf-8",
        ) as file:
            detection = json.load(file)

        # convert the prediction into rbai format for single task
        labels = []
        score_sum = 0
        score_count = 0
        for det in detection["prediction"]:
            (klass, score, x_0, y_0, x_1, y_1) = det
            if score < thresh:
                continue

            score_sum += score
            score_count += 1
            category = taxonomy[int(klass)]
            xnorm = x_0
            ynorm = y_0
            wnorm = x_1 - x_0
            hnorm = y_1 - y_0

            label_entry = {
                "bbox2d": {
                    "xnorm": xnorm,
                    "ynorm": ynorm,
                    "hnorm": hnorm,
                    "wnorm": wnorm,
                },
                "category": [["object", category]],
                "attributes": [],
            }
            labels += [label_entry]

        if score_count == 0:
            score_count = 1
        return_task = {
            "taskId": task["taskId"],
            "priority": 1 - (score_sum / score_count),
            "labels": labels,
        }

        labeled_tasks += [return_task]

    return labeled_tasks


def write_line_to_lst(img_path: str, boxes: np.ndarray, ids: np.ndarray, idx: int):
    """
    Write a single prediction line to .lst file.

    .lst file format: A, B, C, D, E
        A: image index
        B: header length = 2
        C: length of one bbox label = 5
        D: all the bounding box labels
            [class id, xmin, ymin, xmax, ymax]
        E: image path
    """
    header_length = 2
    label_length = 5

    # concat id and bboxes
    labels = np.hstack((ids.reshape(-1, 1), boxes)).astype("float")

    # flatten
    labels = labels.flatten().tolist()
    str_idx = [str(idx)]
    str_header = [str(x) for x in [header_length, label_length]]
    str_labels = [str(x) for x in labels]
    str_path = [img_path]
    line = "\t".join(str_idx + str_header + str_labels + str_path) + "\n"
    return line


def write_image_to_disk(root_dir: str, items: str, presigned_items: str):
    """
    Write image to disk found at presigned_items.

    Parameters
    ------------
    root_dir: str
        root director of path.

    items: str
        Checks if exists within this path, if not creates it.

    presigned_items: str
        The url of the image to download.
    """
    filename = os.path.join(root_dir, items)

    # Create the directory recursively
    os.makedirs(os.path.dirname(filename), exist_ok=True)

    # Get image data, and write to disk
    if not os.path.exists(filename):
        image = imageio.imread(presigned_items)
        imageio.imwrite(filename, image)


def convert_redbrick_to_sagemaker(  # pylint: disable=too-many-locals
    info: Dict, cache_dir: str
) -> Tuple[Dict[str, int], int]:
    """
    Convert RedBrick labels to sagemaker format.

        - The RedBrick AI json object is first converted into .lst files.
          A single entry in the .lst file is defined in write_to_line_lst.

        - The .lst files are converted to .rec files using the im2rec helper
          function. The .rec file contains both image and label data.
    """
    result = info["labeled"]
    _, taxonomy = create_taxonomy_map(info["taxonomy"]["categories"][0]["children"])

    # clean output directories for images
    os.makedirs(os.path.join(cache_dir, "images"), exist_ok=True)

    train_file = os.path.join(cache_dir, "annotation_redbrick_train.lst")
    val_file = os.path.join(cache_dir, "annotation_redbrick_validation.lst")
    if os.path.exists(train_file):
        os.remove(train_file)
    if os.path.exists(val_file):
        os.remove(val_file)

    train_split = 0.8

    train_idx = 0
    test_idx = 0
    no_labels = 0
    for _, task in enumerate(tqdm.tqdm(result)):
        # skip tasks without labels
        if len(task["labels"]) == 0:
            no_labels += 1
            continue

        # required for creating the record io file
        write_image_to_disk(
            os.path.join(cache_dir, "images"),
            task["items"][0],
            task["itemsPresigned"][0],
        )

        # iterate over labels, compute class_id and store labels
        all_boxes = np.array([])
        all_boxes_ids = np.array([])
        for label_idx, label in enumerate(task["labels"]):

            # compute the class_id
            class_ = label["category"][0][-1]
            class_id = 0
            if class_ in taxonomy:
                class_id = taxonomy[class_]
            else:
                raise ValueError("%s is not part of your taxonomy!" % class_)

            # append box to box_array
            box = np.array(
                [
                    label["bbox2d"]["xnorm"],
                    label["bbox2d"]["ynorm"],
                    label["bbox2d"]["xnorm"] + label["bbox2d"]["wnorm"],
                    label["bbox2d"]["ynorm"] + label["bbox2d"]["hnorm"],
                ]
            )
            if label_idx == 0:
                all_boxes = np.array([box])
            else:
                all_boxes = np.vstack([all_boxes, box])

            # store the corresponding class_id's
            all_boxes_ids = np.append(all_boxes_ids, class_id)

        # split data into test and train, save images
        if np.random.uniform() <= train_split:
            train_list_line = write_line_to_lst(
                os.path.join(cache_dir, "images", task["items"][0]),
                all_boxes,
                all_boxes_ids,
                train_idx,
            )
            train_idx += 1

            # write to lst file
            with open(
                os.path.join(cache_dir, "annotation_redbrick_train.lst"),
                "a+",
                encoding="Utf-8",
            ) as file:
                file.write(train_list_line)

        else:
            test_list_line = write_line_to_lst(
                os.path.join(cache_dir, "images", task["items"][0]),
                all_boxes,
                all_boxes_ids,
                test_idx,
            )
            test_idx += 1

            # write to lst file
            with open(
                os.path.join(cache_dir, "annotation_redbrick_validation.lst"),
                "a+",
                encoding="Utf-8",
            ) as file:
                file.write(test_list_line)

    # create record io files using im2rec
    run_im2rec(
        [
            "--resize",
            "256",
            "--pack-label",
            "--prefix",
            ".redbrick_sagemaker/annotation_redbrick",
            "--root",
            ".",
        ]
    )

    return taxonomy, train_idx



"""Main client for redbrick sagemaker."""
import json
import os
from typing import Any, Dict, Optional, Tuple

import boto3
import boto3.session
import sagemaker
import sagemaker.transformer

from sagemaker import get_execution_role
from sagemaker.amazon.amazon_estimator import get_image_uri

from redbrick_sagemaker.utils.sagemaker import (
    construct_training_job,
    construct_tuning_job,
    construct_warm_start_config,
)
from redbrick_sagemaker.logging import logger


class SagemakerClient:
    """Interface to run Active Learning cycles with sagemaker."""

    def __init__(
        self,
        bucket: str,
        bucket_prefix: str,
        iam_role: Optional[str],
        cache_dir: str,
        project_id: str,
        tuning_job: Optional[str] = None,
        profile_name: Optional[str] = "default",
    ) -> None:
        """Construct client."""
        # Define sagemaker execution role
        try:
            self.role: str = get_execution_role()
        except ValueError as error:
            if iam_role:
                self.role = iam_role
            else:
                raise Exception(
                    "When using redbrick_sagemaker outside of AWS notebooks,"
                    + "you have to provide an IAM role"
                ) from error

        # Create sagemaker session
        self.boto_session: boto3.session.Session = boto3.session.Session(
            profile_name=profile_name
        )
        self.sagemaker_session: sagemaker.Session = sagemaker.Session(self.boto_session)

        self.bucket: str = bucket
        self.bucket_prefix: str = bucket_prefix
        self.s3_train: str = f"s3://{bucket}/{bucket_prefix}/train"
        self.s3_validation: str = f"s3://{bucket}/{bucket_prefix}/validation"

        self.cache_dir: str = cache_dir
        self.s3_image_inference = "image_inference"
        self.s3_pred_inference = "pred_inference"
        self.project_id = project_id

        # Unintialized
        self.predictor: Optional[Any] = None
        self.runtime: Optional[Any] = None
        self.tuning_job_name: Optional[str] = tuning_job

    def upload_images_to_s3(
        self,
        image_dir: str,
    ):
        """
        Upload image directory to s3.

        Parameters
        -------------
        image_dir: str
            Sub folder where images are stored.
            They will be uploaded to s3 under the same folder.
        """
        self.sagemaker_session.upload_data(
            path=image_dir,
            bucket=self.bucket,
            key_prefix=f"{self.bucket_prefix}/{self.s3_image_inference}",
        )

    def download_predictions_from_s3(self, inference_dir: str):
        """
        Download predictions from client s3.

        Predictions are stored in folder self.s3_image_inference in client
        s3 bucket. Function will download them to inference_dir.

        Parameters
        -------------
        inference_dir: str
            Local directory in which inferences
            will be downloaded.
        """
        self.sagemaker_session.download_data(
            path=inference_dir,
            bucket=self.bucket,
            key_prefix=f"{self.bucket_prefix}/{self.s3_pred_inference}",
        )

    def upload_to_s3(self) -> None:
        """Upload files to s3 for training and validation."""
        self.sagemaker_session.upload_data(
            path=os.path.join(self.cache_dir, "annotation_redbrick_validation.rec"),
            bucket=self.bucket,
            key_prefix=f"{self.bucket_prefix}/validation",
        )
        self.sagemaker_session.upload_data(
            path=os.path.join(self.cache_dir, "annotation_redbrick_train.rec"),
            bucket=self.bucket,
            key_prefix=f"{self.bucket_prefix}/train",
        )

    def _check_if_previous_job_exists(self) -> Tuple[str, Optional[str]]:
        """
        Check if a previous tuning job exists.

        If a previous tuning job exists for this project_id,
        returns the name of the tuning job.

        Returns
        ----------
        Tuple[Optional[str], str]
            The new tuning jobs name, and the previous tuning job name for this project.
            If there was no tuning job previous to this, return None.
        """
        smclient = self.boto_session.client("sagemaker")
        tuning_jobs = smclient.list_hyper_parameter_tuning_jobs()

        job_name_check = "redbrick-" + str(hash(self.project_id))
        for job in tuning_jobs["HyperParameterTuningJobSummaries"]:
            if job["HyperParameterTuningJobName"].startswith(job_name_check):
                old_job_name = job["HyperParameterTuningJobName"]
                if len(old_job_name) == 29:
                    new_job_name = old_job_name + "(1)"
                else:
                    new_job_name = old_job_name[0:-3] + f"({int(old_job_name[-2]) + 1})"
                return new_job_name, old_job_name

        return job_name_check, None

    def launch_tuning_job(self, total_training_samples: int):
        """Launch a hyperparameter tuning job."""
        training_image = get_image_uri(
            self.sagemaker_session.boto_region_name,
            "object-detection",
            repo_version="latest",
        )

        # Create a tuning job name
        new_job_name, old_job_name = self._check_if_previous_job_exists()
        self.tuning_job_name = new_job_name

        # Construct tuning job config
        tuning_job_config = construct_tuning_job(
            num_training_samples=total_training_samples
        )

        # Construct warmstart config
        warm_start_config = None
        if old_job_name:
            warm_start_config = construct_warm_start_config(old_job_name)

        # Construct training job config
        training_job_config = construct_training_job(
            training_image=training_image,
            train_uri=self.s3_train,
            validation_uri=self.s3_validation,
            bucket=self.bucket,
            num_classes=3,
            num_training_samples=total_training_samples,
            sagemaker_role=self.role,
        )
        smclient = self.boto_session.client("sagemaker")
        smclient.create_hyper_parameter_tuning_job(
            HyperParameterTuningJobName=self.tuning_job_name,
            HyperParameterTuningJobConfig=tuning_job_config,
            TrainingJobDefinition=training_job_config,
            WarmStartConfig=warm_start_config,
        )

    def transform(self):
        """Create a transformer, and start transform."""
        _, model_name = self.best_training_job()
        model = self.sagemaker_session.create_model_from_job(model_name)

        # create the location path for storing predictions,
        # and create transformer
        output_path = (
            f"s3://{self.bucket}/{self.bucket_prefix}/{self.s3_pred_inference}"
        )
        transformer = sagemaker.transformer.Transformer(
            model_name=model,
            instance_count=1,
            instance_type="ml.p2.xlarge",
            output_path=output_path,
        )

        # specify location of images for predictions, and begin transform
        in_path = (
            f"s3://{self.bucket}/" + f"{self.bucket_prefix}/{self.s3_image_inference}"
        )
        transformer.transform(in_path, content_type="image/jpeg", logs=False)

    def best_training_job(self) -> Tuple[bool, str]:
        """Return the best training job."""
        tuner = sagemaker.tuner.HyperparameterTuner.attach(self.tuning_job_name)
        try:
            best_job = tuner.best_training_job()
            return True, best_job
        except Exception as error:  # pylint: disable=broad-except
            return False, str(error)

    def describe(self, log: bool = True) -> Dict:
        """Check the status of the tuning job."""
        tuner = sagemaker.tuner.HyperparameterTuner.attach(self.tuning_job_name)
        describe = tuner.describe()

        if log:
            # Log useful info
            logger.info(
                "INFO: Your tuning job is %s",
                describe["HyperParameterTuningJobStatus"],
            )
            logger.info(
                "INFO: Summary of training jobs: %s",
                json.dumps(describe["TrainingJobStatusCounters"], indent=2),
            )
        return describe

    def deploy_model(self, model_name):
        """Deploy model to endpoint, and init self."""
        estimator = sagemaker.estimator.Estimator.attach(model_name)
        predictor = estimator.deploy(
            initial_instance_count=1, instance_type="ml.m4.xlarge"
        )
        self.predictor = predictor
        self.runtime = boto3.client(service_name="runtime.sagemaker")

    def predict(self, image_bytes: bytearray) -> Dict[Any, Any]:
        """Make a prediction using deployed predictor."""
        if not self.runtime:
            raise Exception("No runtime!")
        if not self.predictor:
            raise Exception("No predictor!")

        endpoint_response = self.runtime.invoke_endpoint(
            EndpointName=self.predictor.endpoint,
            ContentType="image/jpeg",
            Body=image_bytes,
        )
        results = endpoint_response["Body"].read()
        detections = json.loads(results)
        return detections

    def wait_for_tuning(self):
        """Wait for tuning job."""
        tuner = sagemaker.tuner.HyperparameterTuner.attach(self.tuning_job_name)
        tuner.wait()

    def delete_endpoint(self):
        """Delete a deployed endpoint."""
        # No deployed predictor
        if self.predictor is None:
            return

        endpoint_name = self.predictor.endpoint_name
        client = boto3.client("sagemaker")
        client.delete_endpoint(EndpointName=endpoint_name)



"""Setup the python module."""

import os
from setuptools import setup, find_packages  # type: ignore

with open(
    "README.md",
    "r",
    encoding="utf-8",
) as fh:
    long_description = fh.read()

with open(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "redbrick_sagemaker",
        "VERSION",
    ),
    "r",
    encoding="utf-8",
) as f:
    version = f.read().strip()

install_requires = [
    "redbrick-sdk>=0.6.13",
    "sagemaker==2.60.0",
    "boto3==1.20.11",
    "mxnet>=1.5.0",
    "opencv-python==4.5.4.60",
]

setup(
    name="redbrick_sagemaker",
    url="https://github.com/redbrick-ai/redbrick-sagemaker",
    version=version,
    description="RedBrick AI and AWS Sagemaker integration",
    py_modules=["redbrick-sagemaker"],
    python_requires=">=3.7.0, <3.10",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        "dev": [
            "wheel==0.35.1",
            "pytest-mock==3.3.1",
            "pytest==6.1.2",
            "black==21.7b0",
            "mypy==0.790",
            "mypy-extensions==0.4.3",
            "pylint==2.11.1",
            "pycodestyle==2.6.0",
            "pydocstyle==5.1.1",
            "flake8==3.8.4",
            "flake8-print==3.1.4",
            "eradicate==1.0",
            "pytest-cov>=2.8.1",
            "pytest-asyncio==0.12.0",
            "pytest-eradicate==0.0.5",
            "pytest-pycodestyle==2.2.0",
            "pytest-pydocstyle==2.2.0",
            "pytest-black==0.3.12",
            "pytest-pylint==0.17.0",
            "pytest-mypy==0.7.0",
            "pytest-flake8==1.0.6",
            "pytest-randomly==3.4.1",
        ]
    },
)



"""Active Learning."""
from typing import Dict, Optional, Tuple

import json
import os

import redbrick
from redbrick_sagemaker.logging import logger
from redbrick_sagemaker.utils.bbox import (
    convert_redbrick_to_sagemaker,
    convert_sagemaker_rbai_bbox,
)
from redbrick_sagemaker.utils.redbrick import (
    create_taxonomy_map,
)
from redbrick_sagemaker.utils.dataio import download_images_from_remote
from redbrick_sagemaker.sagemaker_client import SagemakerClient


class ActiveLearner:
    """Runs the Active Learning process."""

    def __init__(
        self,
        api_key: str,
        org_id: str,
        project_id: str,
        s3_bucket: str,
        s3_bucket_prefix: str,
        tuning_job: Optional[str] = None,
        url: Optional[str] = "https://api.redbrickai.com",
        iam_role: Optional[str] = None,
    ):
        """Construct object."""
        self.project_id: str = project_id
        self.org_id: str = org_id
        self.api_key: str = api_key

        # get redbrick ai project object
        self.project: redbrick.RBProject = redbrick.get_project(
            api_key, url, org_id, project_id
        )

        # prepare cache directory
        self.cache_dir = ".redbrick_sagemaker"
        os.makedirs(self.cache_dir, exist_ok=True)

        # prepare sagemaker client
        self.sagemaker_client = SagemakerClient(
            bucket=s3_bucket,
            bucket_prefix=s3_bucket_prefix,
            iam_role=iam_role,
            cache_dir=self.cache_dir,
            tuning_job=tuning_job,
            project_id=self.project_id,
        )

        # info for learning and inference
        self.info: Optional[Dict] = None

    def _get_learning_info(self):
        """Get learning info for active learning."""
        info = self.project.learning2.get_learning_info(min_new_tasks=0)
        with open(
            os.path.join(self.cache_dir, "cache__info.json"),
            "w+",
            encoding="Utf-8",
        ) as file:
            json.dump(info, file, indent=2)

        return info

    def _update_info(self):
        """Update info object."""
        # if self.info is None:
        self.info = self._get_learning_info()

    def train(self):
        """Launch a training job."""
        # Get info from RBAI remote - for learning and inference
        logger.info("Begining a hyperparameter optimization job.")

        self._update_info()
        self.project.learning2.start_processing()

        try:
            # Create training files required by sagemaker, upload to s3
            _, total_training_samples = convert_redbrick_to_sagemaker(
                self.info, cache_dir=self.cache_dir
            )

            logger.info("Uploading training files to s3.")
            self.sagemaker_client.upload_to_s3()

            logger.info("Launching tuning job.")
            # Launch a tuning job
            self.sagemaker_client.launch_tuning_job(total_training_samples)

        except Exception as error:
            # clean up
            self.project.learning2.end_processing()
            raise Exception from error

    def inference(self):
        """Perform inference on unlabeled data."""
        try:
            # Get updated unlabeled tasks from rbai
            self._update_info()
            unlabeled_tasks = self.info["unlabeled"]

            # Download images from remote rbai storage
            logger.info("Downloading images from RedBrick AI.")
            download_images_from_remote(
                self.cache_dir, "images_inference", unlabeled_tasks
            )

            # Upload images for inference to client s3 bucket
            logger.info("Uploading images to client s3 bucket for inference.")
            self.sagemaker_client.upload_images_to_s3(
                os.path.join(self.cache_dir, "images_inference")
            )

            # Begin batch transform job to generate predictions
            logger.info(
                "Starting transform operation to generate %s",
                " predictions in a batch.",
            )
            self.sagemaker_client.transform()

            # Download predictions from client s3
            logger.info("Retrieving predictions from client s3.")
            self.sagemaker_client.download_predictions_from_s3(
                os.path.join(self.cache_dir, "pred_inference")
            )

            # Post process predictions, and convert to RedBrick format
            taxonomy_map, _ = create_taxonomy_map(
                self.info["taxonomy"]["categories"][0]["children"]
            )
            labeled_tasks = convert_sagemaker_rbai_bbox(
                prediction_dir=os.path.join(self.cache_dir, "pred_inference"),
                tasks=unlabeled_tasks,
                taxonomy=taxonomy_map,
            )

        except Exception as error:
            # Clean up endpoint
            raise Exception from error

        return labeled_tasks

    def _status(self) -> Tuple[bool, str]:
        """Check status of training job."""
        return self.sagemaker_client.best_training_job()

    def describe(self):
        """Describe status."""
        self.sagemaker_client.describe()

    def end_processing(self):
        """Force stop processing."""
        self.project.learning2.end_processing()

    def start_processing(self):
        """Force start processing."""
        self.project.learning2.start_processing()

    def run(self, wait: bool = False, force_run: bool = False):
        """Run active learning cycle."""
        is_processing = self.project.learning2.check_is_processing()
        # Check if a tuning job is in progress
        if is_processing:
            # Check if the tuning job is complete
            describe = self.sagemaker_client.describe(log=False)
            tuning_complete = describe["HyperParameterTuningJobStatus"] == "Completed"

            # Check if user wants to perform inference with best model
            if force_run or tuning_complete:
                check, _ = self._status()

                if not check:
                    self.sagemaker_client.describe()

                    jobname = self.sagemaker_client.tuning_job_name
                    raise Exception(
                        f"Your tuning job {jobname}"
                        + " doesn't have a best model it can use"
                        + " for inference yet."
                    )

                # Use best model is available for inference
                logger.info("Starting the inference process on unlabeled data.")
                labeled_tasks = self.inference()

                logger.info(
                    "Updating tasks on RedBrick with %s",
                    "predictions and uncertainty scores.",
                )
                self.project.learning2.update_tasks(1, labeled_tasks)
                self.project.learning2.end_processing()

            else:
                logger.warning("An Active learning process is already in progress.")

        else:
            self.train()
            if wait:
                self.sagemaker_client.wait_for_tuning()

                # Perform inference
                check, _ = self._status()
                labeled_tasks = self.inference()
                self.project.learning2.update_tasks(1, labeled_tasks)
                self.project.learning2.end_processing()



"""RedBrick related helper functions."""


def create_taxonomy_map(taxonomy):
    """
    Create a taxonomy map from RBAI taxonomy object.

    Parameters
    -------------
    taxonomy: Dict
        Redbrick AI taxonomy object:
        [
            {
                "name": str,
                "class_id": int,
                "children": []
            }
            .
            .
        ]

    Returns
    --------------
    taxonomy_map: Dict[int, str]
        maps from class_id to category name
    """
    taxonomy_map = {}
    inverted_taxonomy_map = {}
    for category in taxonomy:
        taxonomy_map[category["classId"]] = category["name"]
        inverted_taxonomy_map[category["name"]] = category["classId"]

    return taxonomy_map, inverted_taxonomy_map



"""Logging for redbrick_sagemaker."""

import logging


class Logger:
    """Custom logger."""

    def __init__(self):
        """Construct logging object."""
        # Create a custom logger
        logger_ = logging.getLogger(__name__)
        logger_.setLevel(logging.INFO)

        # Create handlers
        console = logging.StreamHandler()

        # Create formatters and add it to handlers
        console.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))

        # Add handlers to the logger
        logger_.addHandler(console)

        self.logger = logger_


logger = Logger().logger
